
#include "FD2E.h"

/***************************************************************
// Global Variable
***************************************************************/
rb_red_blk_tree* SHA1tree;
rb_red_blk_tree* Rabintree;

void UInt64Dest(void* a);
int UInt64Comp(const void* a,const void* b);
void UInt64Print(const void* a);
int cmp(const void* a, const void* b);
int HashComp(const void* a,const void* b);
int Hashcpy(const void* a,const void* b);
void HashPrint(const void* a);
void InfoPrint(void* a);
void InfoDest(void *a);
void HashDest(void* a);
u_int64_t RabinBlockHash(unsigned char *buf);
u_int64_t RabinVBlockHash(unsigned char *buf, int size);
u_int64_t RabinVBlockHash(unsigned char *buf, int size);
u_int64_t RabinVariableHash(unsigned char *buf, int size);
int FindSHA1(unsigned char* SFP);
int FindRabin(u_int64_t SFP);
void InsertSHA1(unsigned char* SFP);
void InsertRabin(u_int64_t SFP);
void NodupInsertSHA1(unsigned char* SFP);
void NodupInsertRabin(u_int64_t SFP);
void DBC_FS(int fdr);
int SlideVerify(unsigned char *window1, unsigned char *sha1sum_back, int size, int *k, u_int64_t *raf);
int VerifyHash(unsigned char* buf, unsigned char *sha1sum, int size, u_int64_t* ra );
void WriteLog(int fdmw, u_int64_t rabinf, unsigned char * sha1, u_int64_t offset, int size);
int VerifyFixedBlock(u_int64_t *offset, 
				 u_int64_t filesize, 
				 int fdr,
				 int fdmw,
				 unsigned char *buf, 
				 unsigned char *sha1sum,
				 u_int64_t *rabinf,
				 int *chunk_ref_count);
int VerifyStrideBlock(unsigned char *window1,
						unsigned char *buf,
						unsigned char *buf1,
						unsigned char *sha1sum_back,
						unsigned char *sha1sum,
						u_int64_t *rabinp,
						u_int64_t *rabinf,
						u_int64_t *offset,
						u_int64_t filesize,
						int fdmw,
						int fdr,
						int stride);

int VerifyISSFixedBlock(u_int64_t *offset, 
				 u_int64_t filesize, 
				 int fdr,
				 int fdmw,
				 unsigned char *buf, 
				 unsigned char *sha1sum,
				 u_int64_t *rabinf,
				 int *chunk_ref_count,
				 int stride);

int VerifyISSBlock(unsigned char *window1,
						unsigned char *buf,
						unsigned char *buf1,
						unsigned char *sha1sum_back,
						unsigned char *sha1sum,
						u_int64_t *rabinp,
						u_int64_t *rabinf,
						u_int64_t *offset,
						u_int64_t filesize,
						int fdmw,
						int fdr,
						int stride);


int VerifyTailBlock(unsigned char *window1,
						unsigned char *sha1sum_back,
						unsigned char *sha1sum,
						unsigned char *buf,
						unsigned char *buf1,
						u_int64_t *rabinp,
						u_int64_t *rabinf,
						u_int64_t *offset,
						u_int64_t filesize,
						int fdmw,
						int fdr 
						);
void VerifyBackTracking(u_int64_t offset,
						u_int64_t filesize,
						u_int64_t localtemp,
						int fdmw,
						int fdr,
						int stride
						);

void DBC_ISS(int fdr, int stride);
void DBC_SS(int fdr, int stride);
void DBC_BS(int fdr);

window myRabin(FINGERPRINT_PT, CHUNK_SIZE);
int Chunk_Count;
struct timeval start, end;

/***************************************************************
// Free Allocation
***************************************************************/
void UInt64Dest(void* a) {
	free((u_int64_t*)a);
}

/***************************************************************
// Uint64 Compare Function
***************************************************************/
int UInt64Comp(const void* a,const void* b) {
	if( *(u_int64_t*)a > *(u_int64_t*)b) return(1);
	if( *(u_int64_t*)a < *(u_int64_t*)b) return(-1);
	return(0);
}

/***************************************************************
// Uint64 Print
***************************************************************/
void UInt64Print(const void* a) {
	printf("%llu\n",*(u_int64_t*)a);
}


/***************************************************************
// Sha-1 hash Compare
// if return value is 0 then hash is equal
***************************************************************/
int cmp(const void* a, const void* b)
{
	for(int i=0; i<20; i+=1)
	{
		if( *((char *)a+i) > *((char *)b+i))
		{
			return 1;
		}
		else if( *((char *)a+i) < *((char *)b+i))
		{
			return -1;
		}
	}

	return 0;
}

/***************************************************************
// SHA1 Hash Value free allocation
***************************************************************/
void HashDest(void* a) {
	int i=0;
	for(i=0; i<20; i++)
	free((char*)a+i);
}

/***************************************************************
// SHA1 Hash Value free allocation
***************************************************************/
int HashComp(const void* a,const void* b) {
	if( cmp(a,b) > 0 ) return(1);
	if( cmp(a,b) < 0 ) return(-1);
	return(0);
}

/***************************************************************
// SHA1 Hash Value Copy
***************************************************************/
int Hashcpy(const void* a,const void* b) {
	int i = 0;
	for(i=0; i<20; i++)
	{
		*((unsigned char *)a+i) = *((unsigned char *)b+i);
	}
	return 0;
}

/***************************************************************
// SHA1 Hash - Hex Decimal Print
***************************************************************/
void HashPrint(const void* a) {
	int i=0;
	unsigned int k;
	
	for(i=0; i<20; i++)
	{
		k=*((unsigned char *)a+i) >> 4;
		printf("%X", k);
		k=*((unsigned char *)a+i) & 0x0F;
		printf("%X", k);
		k=0;
	}

	printf("\n");
}

/***************************************************************
// rbtree basis function
***************************************************************/
void InfoPrint(void* a) {
	;
}

/***************************************************************
// rbtree basis function
***************************************************************/
void InfoDest(void *a){
	;
}

/***************************************************************
// Rabin Block hash Function -   Ʈ
//  ΰƮ  üũ ´.
//    
***************************************************************/
u_int64_t RabinBlockHash(unsigned char *buf)
{
	u_int64_t rabinf = 0;

	window myR(FINGERPRINT_PT, CHUNK_SIZE);	// INITIALIZE RABINPOLY CLASS
	myR.reset(); // RESET if you want to clear up the buffer in the RABINPOLY CLASS

	for(int i=0; i<CHUNK_SIZE; i++)
	{
		rabinf = myR.slide8(buf[i]);
	}

	return rabinf;
}


/***************************************************************
// Rabin Block hash Function -   Ʈ
//  ΰƮ  üũ ´.
//  ũ⸦ ־ Ӽ ִ ؽ  
***************************************************************/
u_int64_t RabinVBlockHash(unsigned char *buf, int size)
{
	u_int64_t rabinf = 0;

	myRabin.reset();
	for(int i=0; i<size; i++)
	{
	        rabinf = myRabin.slide8(buf[i]);
	}

	return rabinf;
}

/***************************************************************
// Rabin Block hash Function -   Ʈ
//  ΰƮ  üũ ´.
//  ũ⸦ ־ Ӽ ִ ؽ  
***************************************************************/
u_int64_t RabinVariableHash(unsigned char *buf, int size)
{
	u_int64_t rabinf = 0;

	window myR(FINGERPRINT_PT, size);	// INITIALIZE RABINPOLY CLASS
	myR.reset(); // RESET if you want to clear up the buffer in the RABINPOLY CLASS

	for(int i=0; i<size; i++)
	{
	        rabinf = myR.slide8(buf[i]);
	}

	return rabinf;
}

/***************************************************************
// find SHA-1
// returning plus digit is finding hash
// returning minus digit is no finding hash
***************************************************************/
int FindSHA1(unsigned char* SFP)
{
	rb_red_blk_node* newNode;
	if( (newNode = RBExactQuery(SHA1tree, SFP)) != NULL )
		return 1000;
	else
		return -500;
}

/***************************************************************
// find Rabin hash
// returning plus digit is finding hash
// returning minus digit is no finding hash
***************************************************************/
int FindRabin(u_int64_t SFP)
{
	rb_red_blk_node* newNode;
	if((newNode = RBExactQuery(Rabintree, &SFP)) != NULL )
		return 1000;
	else
		return -500;
}

/***************************************************************
// Input SHA1 hash into SHA1 Hash Tree
// SHA1 Hash Tree is RB-Tree
// SHA1 is byte stream
***************************************************************/
void InsertSHA1(unsigned char* SFP)
{
	RBTreeInsert(SHA1tree, SFP, 0);
}

/***************************************************************
// Input Rabin hash into Rabin Hash Tree
// Rabin Hash Tree is RB-Tree
// Rabin hash is 64Bit INT
***************************************************************/
void InsertRabin(u_int64_t SFP)
{
	u_int64_t* newKey = (u_int64_t *)malloc(sizeof(u_int64_t));
	*newKey = SFP;
	RBTreeInsert(Rabintree, newKey, 0);
}

void NodupInsertSHA1(unsigned char* SFP)
{
	if( FindSHA1(SFP) < 0 )
	{
		RBTreeInsert(SHA1tree, SFP, 0);
	}
}

void NodupInsertRabin(u_int64_t SFP)
{
	if( FindRabin(SFP)< 0 )
	{
		u_int64_t* newKey = (u_int64_t *)malloc(sizeof(u_int64_t));
		*newKey = SFP;
		RBTreeInsert(Rabintree, newKey, 0);
	}
}


/***************************************************************
// Fixed Size De-Duplication
// ߿ ߺ  ˰ 
***************************************************************/
void DBC_FS(int fdr)
{
	// local variable
	int i=0;
	int readLen = 0;
	int fdmw = open64("./READMETA", O_RDWR | O_TRUNC | O_CREAT);
	unsigned int k=0;
	char metabuf[100];
	char metabuf1[300];
	unsigned char buf[CHUNK_SIZE];
	u_int64_t filesize = lseek64(fdr, 0, SEEK_END);
	u_int64_t offset = 0;

	Chunk_Count = 0;
	lseek64(fdr, 0, SEEK_SET);

	while( offset < filesize )
	{
		// ؽ 
		unsigned char *sha1sum;
		sha1sum = (unsigned char*)malloc(20);
		memset(buf, '\0', CHUNK_SIZE);
		readLen = read(fdr, buf, CHUNK_SIZE);
		make_sha1_checksum(buf, readLen, sha1sum);

		// Ÿ  
		memset(metabuf, '\0', 100);
		memset(metabuf1, '\0', 300);
		for(i=0; i<20; i++)
		{
			k=*((unsigned char *)sha1sum+i) >> 4;
			sprintf(&metabuf[2*i], "%X", k);
			k=*((unsigned char *)sha1sum+i) & 0x0F;
			sprintf(&metabuf[2*i+1], "%X", k);
			k=0;
		}

		if( FindSHA1(sha1sum) < 0 )
		{
			// ߺ  
			InsertSHA1(sha1sum);
			sprintf(metabuf1, "%s:%llu:%d:1\n", metabuf, offset, readLen);
			offset += readLen;
			write(fdmw, metabuf1, strlen(metabuf1));
		}
		else
		{
			// ߺ  ƴ 
			sprintf(metabuf1, "%s:%llu:%d:0\n", metabuf, offset, readLen);
			offset += readLen;
			write(fdmw, metabuf1, strlen(metabuf1));
			Chunk_Count += 1;
		}
	}

	close(fdmw);
}

/***************************************************************
@brief			 Լ ԷµǴ  window   
				ؽϿ ߺ  ϴ ˻Ѵ.
@date			2009-07-22
@author			chorogyi@hallym.ac.kr
@warning		ȭ 
@return value	 => ߺǴ  
				 => ߺǴ   
				ڰ u_int64_t *raf call-by-reference
				ڰ unsigend char *sha1sum_back call-by-reference
				ڰ k ã ġ
***************************************************************/
int SlideVerify(unsigned char *window1, unsigned char *sha1sum_back, int size, int *k, u_int64_t *raf)
{
	int i = 0;
	u_int64_t rabinp = 0;

	for(i=CHUNK_SIZE; i<CHUNK_SIZE+size; i++)
	{
		rabinp = myRabin.slide8((char)window1[i]);

		if ( FindRabin(rabinp) > 0 )
		{
			// find rabin fp;
			// printf("Same Finger Print\n");
			make_sha1_checksum(window1+i-CHUNK_SIZE+1, CHUNK_SIZE, sha1sum_back);

			if ( FindSHA1(sha1sum_back) > 0 )
			{
				// find duplication block
				*k = i-CHUNK_SIZE;
				*raf = rabinp;
				return -1000;
			}
		}
	}
	
	*raf = rabinp;
	return 1000;
}

/***************************************************************
@brief			 Լ ԷµǴ  ΰϿ 
				 ؽ  RBTree ϴ ˻ϰ 
				   Ѵٸ  SHA1ؽ Ͽ 
				 ϴ ˻Ѵ. 
				 SHA1  ִٸ ߺ  

@date			2009-07-22
@author			chorogyi@hallym.ac.kr
@warning		ȭ 
@return value	 => ߺǴ  
				 => ߺǴ   
				ڰ u_int64_t *ra call-by-reference
***************************************************************/
int VerifyHash(unsigned char* buf, unsigned char *sha1sum, int size, u_int64_t* ra )
{
	*ra = RabinVBlockHash(buf, size);
	if( FindRabin(*ra) > 0 )
	{
		// find rabinfp
		make_sha1_checksum((unsigned char*)buf, size, sha1sum);

		if ( FindSHA1(sha1sum) > 0 )
		{
			// find duplication block
			return -1000;
		}
	}

	// no duplication block
	return 1000;
}

/***************************************************************
@brief			SHA1  RabinHash  α׷ 
@date			2009-07-22
@author			chorogyi@hallym.ac.kr
@warning		ȭ 
@return value	fdmw => log file descriptor
				sha1 => sha1 value
				rabinf => rabinhash value
				offset => the position in a file
				size => block size
***************************************************************/
void WriteLog(int fdmw, u_int64_t rabinf, unsigned char * sha1, u_int64_t offset, int size)
{
	// save infor
	char metabuf[100];
	unsigned int k = 0;
	memset(metabuf, '\0', 100);
	for(int i=0; i<20; i++)
	{
		k=*((unsigned char *)sha1+i) >> 4;
		sprintf(&metabuf[2*i], "%X", k);
		k=*((unsigned char *)sha1+i) & 0x0F;
		sprintf(&metabuf[2*i+1], "%X", k);
		k=0;
	}
	sprintf(metabuf, "%s:%llu:%d:%llu\n", metabuf, offset, size, rabinf);
	write(fdmw, metabuf, strlen(metabuf));
}

/***************************************************************
//  Ǵ  continue ؾϴ, exit ؾϴ, 
// ؾߴ ǥ  ũ ߺ ó
***************************************************************/
int VerifyFixedBlock(u_int64_t *offset, 
				 u_int64_t filesize, 
				 int fdr,
				 int fdmw,
				 unsigned char *buf, 
				 unsigned char *sha1sum,
				 u_int64_t *rabinf,
				 int *chunk_ref_count)
{
	//    ߴ ˻
	if( *offset + CHUNK_SIZE < filesize )
	{
		// CHUNK_SIZE ŭ д´.
		read( fdr, buf, CHUNK_SIZE );
		*offset += CHUNK_SIZE;

		//  ũ  ˻Ͽ  Ǵ  0  ߺ̶ Ǵ
		if (  VerifyHash( buf, sha1sum, CHUNK_SIZE, rabinf ) < 0 )
		{
			// ߺ  ó (Ÿ ͷ )
			//WriteLog(fdmw, *rabinf, sha1sum, (int)(*offset - CHUNK_SIZE), CHUNK_SIZE);
			*chunk_ref_count += 1;

			// ߺ  ߰ ߱   ũ  ˻Ѵ.
			return 1;
		}
		else
		{
			return 0;
		}
	}
	else
	{
		//   ˻  ߺ ˻ϰ ߺ˻  
		read( fdr, buf, (int)(filesize - *offset));

		if ( VerifyHash( buf, sha1sum, CHUNK_SIZE, rabinf ) < 0 )
		{
			//WriteLog(fdmw, *rabinf, sha1sum, *offset, (int)(filesize - *offset));
			*chunk_ref_count += 1;
		}
		else
		{
			// ߺ  ƴҰ   
			unsigned char *sha1;
			sha1 = (unsigned char*)malloc(20);
			make_sha1_checksum(buf, (unsigned long)(filesize-*offset), sha1);

			WriteLog(fdmw, *rabinf, sha1, *offset, (int)(filesize - *offset));
			InsertRabin(*rabinf);
			InsertSHA1(sha1);
		}
		
		/////////////////////////////
		// Loop Exit
		/////////////////////////////
		return -1;
	}
}

/***************************************************************
//  Ǵ  continue ؾϴ, exit ؾϴ, 
// ؾߴ ǥ
***************************************************************/
int VerifyStrideBlock(unsigned char *window1,
						unsigned char *buf,
						unsigned char *buf1,
						unsigned char *sha1sum_back,
						unsigned char *sha1sum,
						u_int64_t *rabinp,
						u_int64_t *rabinf,
						u_int64_t *offset,
						u_int64_t filesize,
						int fdmw,
						int fdr,
						int stride)
{
	int temp = 0;
	u_int64_t readLen=0;

	// window1  ޺κ ̵ ؽõ ˻ϰ  Ǵ  0  ߺ ߰
	if ( SlideVerify(window1, sha1sum_back, CHUNK_SIZE, &temp, rabinp) < 0 )
	{
		// ߺ  ܿ պκ ¥  ߺ ˻Ѵ.
		if ( VerifyHash(window1, sha1sum, temp, rabinf) > 0 )
		{
			//  ߺ  ƴ϶ ο  Ѵ.
			InsertRabin(*rabinf);
			unsigned char *sha1;
			sha1 = (unsigned char*)malloc(20);
			make_sha1_checksum(window1, temp, sha1);
			InsertSHA1(sha1);
			WriteLog(fdmw, *rabinf, sha1, *offset-CHUNK_SIZE, temp);
		}
		else
		{
			//  ߺ ̶ Ÿ ͸ 
			//WriteLog(fdmw, *rabinf, sha1sum, *offset-CHUNK_SIZE, temp);
			Chunk_Count += 1;
		}

		Chunk_Count += 1;
		*offset = *offset + temp + 1;
		//WriteLog(fdmw, *rabinp, sha1sum_back, *offset - CHUNK_SIZE, CHUNK_SIZE);
		lseek(fdr, *offset, SEEK_SET);
		
		return 1;
	}
	else
	{
		//ߺ ߰  ߱  Stride ϱ
		unsigned char *sha1;
		sha1 = (unsigned char*)malloc(20);
		
		// [ f ]|[ p ] f  Ʈ  ʾұ  ߺ ٽ Ȯ ؾ
		if( *rabinf == *rabinp )
		{
			make_sha1_checksum(buf, CHUNK_SIZE, sha1sum);
			make_sha1_checksum(buf1, CHUNK_SIZE, sha1sum_back);

			// SHA1 ؽ 
			if( HashComp((const void *)sha1sum, (const void *)sha1sum_back) == 0)
			{
				memcpy(buf, buf1, CHUNK_SIZE);
				Hashcpy((const void *)sha1, (const void *)sha1sum);
				InsertRabin(*rabinf);
				InsertSHA1(sha1);
				*rabinf = *rabinp;

				// f 
				WriteLog(fdmw, *rabinf, sha1, *offset-CHUNK_SIZE, CHUNK_SIZE);

				//  
				Chunk_Count += 1;
				lseek(fdr, *offset, SEEK_SET);
				
				return 1;
			}
		}

		unsigned char *sha2;
		sha2 = (unsigned char*)malloc(20);
		
		// ٸٸ [ f ]  ؾ 
		make_sha1_checksum(buf, CHUNK_SIZE, sha1);
		InsertRabin(*rabinf);
		InsertSHA1(sha1);
		WriteLog(fdmw, *rabinf, sha1, *offset-CHUNK_SIZE, CHUNK_SIZE);

		// [ p ] ϶ ؾ 
		make_sha1_checksum(buf1, CHUNK_SIZE, sha2);
		InsertRabin(*rabinp);
		InsertSHA1(sha2);
		WriteLog(fdmw, *rabinp, sha2, *offset, CHUNK_SIZE);
		*offset += CHUNK_SIZE;
			
		// Stride Step
		if( stride < 3 )
		{
			return 1;
		}

		int i;
		for(i = 0; i < stride - 2; i++)
		{
			unsigned char *sha3;
			sha3 = (unsigned char*)malloc(20);
		
			memset(buf, '\0', CHUNK_SIZE);
			readLen = filesize - *offset;
				
			if(readLen < CHUNK_SIZE)
			{
				read(fdr, buf, readLen);
				make_sha1_checksum(buf, (unsigned long)readLen, sha3);
				*rabinf = RabinVBlockHash(buf, (int)readLen);
			}
			else
			{
				readLen = CHUNK_SIZE;
				read(fdr, buf, CHUNK_SIZE);
				make_sha1_checksum(buf, CHUNK_SIZE, sha3);
				*rabinf = RabinVBlockHash(buf, CHUNK_SIZE);
			}

			if( FindRabin(*rabinf) > 0)
			{
				if( FindSHA1(sha1sum) > 0)
				{
					//WriteLog(fdmw, *rabinf, sha1sum, *offset-readLen, readLen);
					Chunk_Count += 1;
					continue;
				}
			}

			// Ȱ ٸ
			InsertRabin(*rabinf);
			InsertSHA1(sha3);
			WriteLog(fdmw, *rabinf, sha3, *offset, readLen);
			*offset += readLen;
			if( filesize - *offset <= CHUNK_SIZE)
			{
				lseek(fdr, *offset, SEEK_SET);
				return 1;
			}
		}

		return 1;
	}
}

/***************************************************************
//  Ǵ  continue ؾϴ, exit ؾϴ, 
// ؾߴ ǥ  ũ ߺ ó
***************************************************************/
int VerifyISSFixedBlock(u_int64_t *offset, 
				 u_int64_t filesize, 
				 int fdr,
				 int fdmw,
				 unsigned char *buf, 
				 unsigned char *sha1sum,
				 u_int64_t *rabinf,
				 int *chunk_ref_count,
				 int stride)
{
	u_int64_t readLen = 0;
	
	//    ߴ ˻
	if( *offset + CHUNK_SIZE*stride < filesize )
	{
		// CHUNK_SIZE ŭ д´.
		lseek(fdr, *offset, SEEK_SET);
		read( fdr, buf, CHUNK_SIZE );

		//  ũ  ˻Ͽ  Ǵ  0  ߺ̶ Ǵ
		if (  VerifyHash( buf, sha1sum, CHUNK_SIZE, rabinf ) < 0 )
		{
			// ߺ  ó (Ÿ ͷ )
			//WriteLog(fdmw, *rabinf, sha1sum, (int)(*offset - CHUNK_SIZE), CHUNK_SIZE);
			*chunk_ref_count += 1;

			// ߺ  ߰ ߱   ũ  ˻Ѵ.
			return 1;
		}
		else
		{
			return 0;
		}
	}
	else
	{
		//   ˻  ߺ ˻ϰ ߺ˻  
		int i;
		for(i = 0; i < stride; i++)
		{
			unsigned char *sha3;
			sha3 = (unsigned char*)malloc(20);
		
			memset(buf, '\0', CHUNK_SIZE);
			readLen = filesize - *offset;
				
			if(readLen < CHUNK_SIZE)
			{
				read(fdr, buf, readLen);
				*offset += readLen;
				make_sha1_checksum(buf, (unsigned long)readLen, sha3);
				*rabinf = RabinVBlockHash(buf, (int)readLen);
			}
			else
			{
				readLen = CHUNK_SIZE;
				read(fdr, buf, CHUNK_SIZE);
				*offset += CHUNK_SIZE;
				make_sha1_checksum(buf, CHUNK_SIZE, sha3);
				*rabinf = RabinVBlockHash(buf, CHUNK_SIZE);
			}

			if( FindRabin(*rabinf) > 0)
			{
				if( FindSHA1(sha1sum) > 0)
				{
					//WriteLog(fdmw, *rabinf, sha1sum, *offset-readLen, readLen);
					Chunk_Count += 1;
					continue;
				}
			}

			// Ȱ ٸ
			InsertRabin(*rabinf);
			InsertSHA1(sha3);
			WriteLog(fdmw, *rabinf, sha3, *offset, readLen);
			
			if( filesize - *offset <= CHUNK_SIZE)
			{
				lseek(fdr, *offset, SEEK_SET);
				return -1;
			}
		}

		return -1;
	}
}

/***************************************************************
//  Ǵ  continue ؾϴ, exit ؾϴ, 
// ؾߴ ǥ
***************************************************************/
int VerifyISSBlock(unsigned char *window1,
						unsigned char *buf,
						unsigned char *buf1,
						unsigned char *sha1sum_back,
						unsigned char *sha1sum,
						u_int64_t *rabinp,
						u_int64_t *rabinf,
						u_int64_t *offset,
						u_int64_t filesize,
						int fdmw,
						int fdr,
						int stride)
{
	int temp = 0;

	// window1  ޺κ ̵ ؽõ ˻ϰ  Ǵ  0  ߺ ߰
	if ( SlideVerify(window1, sha1sum_back, CHUNK_SIZE, &temp, rabinp) < 0 )
	{
		Chunk_Count += 1;
		//WriteLog(fdmw, *rabinp, sha1sum_back, *offset - CHUNK_SIZE, CHUNK_SIZE);
		
		return temp;
	}
	else
	{
		//ߺ ߰  ߱  Stride ϱ
		unsigned char *sha1;
		sha1 = (unsigned char*)malloc(20);
		
		// [ f ]|[ p ] f  Ʈ  ʾұ  ߺ ٽ Ȯ ؾ
		if( *rabinf == *rabinp )
		{
			make_sha1_checksum(buf, CHUNK_SIZE, sha1sum);
			make_sha1_checksum(buf1, CHUNK_SIZE, sha1sum_back);

			// SHA1 ؽ 
			if( HashComp((const void *)sha1sum, (const void *)sha1sum_back) == 0)
			{
				memcpy(buf, buf1, CHUNK_SIZE);
				Hashcpy((const void *)sha1, (const void *)sha1sum);
				InsertRabin(*rabinf);
				InsertSHA1(sha1);
				*rabinf = *rabinp;

				// f 
				WriteLog(fdmw, *rabinf, sha1, *offset, CHUNK_SIZE);

				//  
				Chunk_Count += 1;
				return 1;
			}
		}

		unsigned char *sha2;
		sha2 = (unsigned char*)malloc(20);
		
		// ٸٸ [ f ]  ؾ 
		make_sha1_checksum(buf, CHUNK_SIZE, sha1);
		InsertRabin(*rabinf);
		InsertSHA1(sha1);
		WriteLog(fdmw, *rabinf, sha1, *offset, CHUNK_SIZE);

		// [ p ] ϶ ؾ 
		make_sha1_checksum(buf1, CHUNK_SIZE, sha2);
		InsertRabin(*rabinp);
		InsertSHA1(sha2);
		WriteLog(fdmw, *rabinp, sha2, *offset+CHUNK_SIZE, CHUNK_SIZE);
			
		return 1;
	}
}

/***************************************************************
//   ˻
***************************************************************/
int VerifyTailBlock(unsigned char *window1,
						unsigned char *sha1sum_back,
						unsigned char *sha1sum,
						unsigned char *buf,
						unsigned char *buf1,
						u_int64_t *rabinp,
						u_int64_t *rabinf,
						u_int64_t *offset,
						u_int64_t filesize,
						int fdmw,
						int fdr 
						)
{
	int temp = 0;
	int readLen = 0;

	//  κ ó
	if ( SlideVerify(window1, sha1sum_back, readLen+1, &temp, rabinp) < 0 )
	{
		//ߺ ߻Ѵٸ
		if ( VerifyHash(window1, sha1sum, temp, rabinf) > 0 )
		{
			//ߺ ߻  պκ ¥ κ üũ
			InsertRabin(*rabinf);
			unsigned char *sha1;
			sha1 = (unsigned char*)malloc(20);
			make_sha1_checksum(window1, temp, sha1);
			InsertSHA1(sha1);
			WriteLog(fdmw, *rabinf, sha1, *offset-CHUNK_SIZE, temp);
		}
		else // ¥ κе ߺ̶
		{
			//WriteLog(fdmw, *rabinp, sha1sum, *offset-CHUNK_SIZE, temp);
			Chunk_Count += 1;
		}

		lseek(fdr, *offset, SEEK_SET);
		Chunk_Count += 1;
		//WriteLog(fdmw, *rabinp, sha1sum, *offset-CHUNK_SIZE, temp);
		return 1;
	}
	else
	{
		// ߺ ߻ ʾ 
		*rabinf = RabinVBlockHash((window1), CHUNK_SIZE);
		InsertRabin(*rabinf);
		unsigned char *sha1;
		sha1 = (unsigned char*)malloc(20);
		make_sha1_checksum(window1, CHUNK_SIZE, sha1);
		InsertSHA1(sha1);
		WriteLog(fdmw, *rabinf, sha1, *offset-CHUNK_SIZE, CHUNK_SIZE);

		*rabinp = RabinVBlockHash((window1+CHUNK_SIZE), readLen);
		InsertRabin(*rabinp);
		unsigned char *sha2;
		sha2 = (unsigned char*)malloc(20);
		make_sha1_checksum(window1+CHUNK_SIZE, readLen, sha2);
		InsertSHA1(sha2);
		WriteLog(fdmw, *rabinp, sha2, *offset, readLen);
		*offset += readLen;
		return -1;
	}
}

/***************************************************************
// ߺ ǽɵǴ κ ˻ // ̿ϼ
***************************************************************/
void VerifyBackTracking(u_int64_t offset,
						u_int64_t filesize,
						u_int64_t localtemp,
						int fdmw,
						int fdr,
						int stride
						)
{
	
	int readLen = 0;
	unsigned char buf[CHUNK_SIZE];
	unsigned char window1[CHUNK_SIZE*2];
	u_int64_t rabinf;

	lseek(fdr, localtemp, SEEK_SET);
	memset(buf, '0', CHUNK_SIZE); //  ʱȭ
	read(fdr, buf, CHUNK_SIZE);
	memcpy(window1, buf, CHUNK_SIZE);

	/*
	while(offset > localtemp)
	{
		memset(buf1, '0', CHUNK_SIZE); //  ʱȭ
		read(fdr, buf1, CHUNK_SIZE)
		memcpy(window1+CHUNK_SIZE, buf1, CHUNK_SIZE);

		unsigned char *sha1;
		sha1 = (unsigned char*)malloc(20);
		
		if ( SlideVerify(window1, sha1, CHUNK_SIZE, &temp, &rabinf) < 0 )
		{

		}
		else
		{
			localtemp += CHUNK_SIZE;
			lseek(fdr, localtemp, SEEK_SET);
			memcpy(window1, buf1, CHUNK_SIZE);
		}
	}
	*/
		
	unsigned char *sha1;
	sha1 = (unsigned char*)malloc(20);

	memset(buf, '\0', CHUNK_SIZE);
	readLen = offset - localtemp;

	if( readLen < CHUNK_SIZE )
	{
		lseek(fdr, localtemp, SEEK_SET);
		read(fdr, buf, readLen);
		offset -= readLen;
		make_sha1_checksum(buf, (unsigned long)readLen, sha1);
		rabinf = RabinVBlockHash(buf, (int)readLen);
	}
	else
	{
		offset -= CHUNK_SIZE;
		lseek(fdr, offset, SEEK_SET);
		readLen = CHUNK_SIZE;
		read(fdr, buf, CHUNK_SIZE);
		make_sha1_checksum(buf, CHUNK_SIZE, sha1);
		rabinf = RabinVBlockHash(buf, CHUNK_SIZE);
	}

	if( FindRabin(rabinf) > 0)
	{
		if( FindSHA1(sha1) > 0)
		{
			//WriteLog(fdmw, *rabinf, sha1sum, *offset-readLen, readLen);
			Chunk_Count += 1;
			//continue;
		}
	}

	// Ȱ ٸ
	InsertRabin(rabinf);
	InsertSHA1(sha1);
	WriteLog(fdmw, rabinf, sha1, offset, readLen);
}

/***************************************************************
// DBC_ISS ߺ  ˻
***************************************************************/
void DBC_ISS(int fdr, int stride)
{
	// local variable
	int state = 0;
	
	int fdmw = open64("./READMETA", O_RDWR | O_TRUNC | O_CREAT);
	
	unsigned char sha1sum[20];
	unsigned char sha1sum_back[20];
	unsigned char buf[CHUNK_SIZE];
	unsigned char buf1[CHUNK_SIZE];
	unsigned char window1[CHUNK_SIZE+CHUNK_SIZE];
	
	u_int64_t filesize = lseek64(fdr, 0, SEEK_END);
	u_int64_t offset = 0;
	u_int64_t rabinf = 0;
	u_int64_t rabinp = 0;

	Chunk_Count = 0;
	
	lseek64(fdr, 0, SEEK_SET);
	
	myRabin.reset(); // RESET if you want to clear up the buffer in the RABINPOLY CLASS

	while(1)
	{
		memset(buf, '0', CHUNK_SIZE);

		// Fixed ϰ  ˻
		state = VerifyFixedBlock(&offset, filesize, fdr, fdmw, buf, sha1sum, &rabinf, &Chunk_Count);

		if( state == -1 ) {
			break;
		}else if( state == 1 ) {
			continue;
		}

		offset -= CHUNK_SIZE;
		offset += CHUNK_SIZE*(stride-2);
		memset(buf, '0', CHUNK_SIZE);
		// Fixed ϰ  ˻  ߺ ߰ߵ  stride  ŭ ǳʺκ ˻
		state = VerifyISSFixedBlock(&offset, filesize, fdr, fdmw, buf, sha1sum, &rabinf, &Chunk_Count, stride);

		if( state == -1 ) {
			break;
		}else if( state == 1 ) {
			// Լ ϳ ʿ
			// localtemp  
			u_int64_t localtemp = offset-(stride-2)*CHUNK_SIZE;
			//  ϵ ˻
			VerifyBackTracking(offset, filesize, localtemp, fdmw, fdr, stride);
			offset += CHUNK_SIZE;
			lseek(fdr, offset, SEEK_SET);
			continue;
		}
		
		memset(buf1, '0', CHUNK_SIZE); //  ʱȭ

		lseek(fdr, offset+CHUNK_SIZE, SEEK_SET);
		read( fdr, buf1, CHUNK_SIZE);
		memcpy(window1, buf, CHUNK_SIZE);
		memcpy(window1+CHUNK_SIZE, buf1, CHUNK_SIZE);

		// ߺ ˻ϰ ߺ  ó
		state = VerifyISSBlock(window1, buf, buf1, sha1sum_back, sha1sum, &rabinp, &rabinf, &offset, filesize, fdmw, fdr, stride );

		if( state == -1 ) {
			break;
		}else if( state == 1 ) {
			u_int64_t localtemp = offset-(stride-2)*CHUNK_SIZE;
			VerifyBackTracking(offset, filesize, localtemp, fdmw, fdr, stride);
			offset += 2*CHUNK_SIZE;
			lseek(fdr, offset, SEEK_SET);
			continue;
		}
		else
		{
			u_int64_t localtemp = offset-(stride-2)*CHUNK_SIZE;
			offset = offset + state;			
			VerifyBackTracking(offset, filesize, localtemp, fdmw, fdr, stride);
			offset += (CHUNK_SIZE + 1);
			lseek(fdr, offset, SEEK_SET);
			continue;
		}
	}
}

/***************************************************************
// Stide shift de-duplication
***************************************************************/
void DBC_SS(int fdr, int stride)
{
	// local variable
	int readLen = 0;
	int state = 0;
	
	int fdmw = open64("./READMETA", O_RDWR | O_TRUNC | O_CREAT);
	
	unsigned char sha1sum[20];
	unsigned char sha1sum_back[20];
	unsigned char buf[CHUNK_SIZE];
	unsigned char buf1[CHUNK_SIZE];
	unsigned char window1[CHUNK_SIZE+CHUNK_SIZE];
	
	u_int64_t filesize = lseek64(fdr, 0, SEEK_END);
	u_int64_t offset = 0;
	u_int64_t rabinf = 0;
	u_int64_t rabinp = 0;

	Chunk_Count = 0;
	
	lseek64(fdr, 0, SEEK_SET);
	
	myRabin.reset(); // RESET if you want to clear up the buffer in the RABINPOLY CLASS

	while(1)
	{
		memset(buf, '0', CHUNK_SIZE);
		state = VerifyFixedBlock(&offset, filesize, fdr, fdmw, buf, sha1sum, &rabinf, &Chunk_Count);

		if( state == -1 ) {
			break;
		}else if( state == 1 ) {
			continue;
		}

		memset(buf1, '0', CHUNK_SIZE); //  ʱȭ
		
		//   κ ˻
		if( offset + CHUNK_SIZE <= filesize )
		{
			read( fdr, buf1, CHUNK_SIZE);
			memcpy(window1, buf, CHUNK_SIZE);
			memcpy(window1+CHUNK_SIZE, buf1, CHUNK_SIZE);

			// ߺ ˻ϰ ߺ  ó
			state = VerifyStrideBlock(window1, buf, buf1, sha1sum_back, sha1sum, &rabinp, &rabinf, &offset, filesize, fdmw, fdr, stride );

			if( state == -1 ) {
				break;
			}else if( state == 1 ) {
				continue;
			}
		}
		else  // ̶
		{
			readLen = filesize - offset;
			read( fdr, buf1, readLen);
			memcpy(window1, buf, CHUNK_SIZE);
			memcpy(window1+CHUNK_SIZE, buf1, readLen);

			state = VerifyTailBlock(window1, buf, buf1, sha1sum_back, sha1sum, &rabinp, &rabinf, &offset, filesize, fdmw, fdr);

			if( state == -1 ) {
				break;
			}else if( state == 1 ) {
				continue;
			}
		}
	}
}

/***************************************************************
@brief			Byte Shift Scheme De-duplication
@date			2009-08-14
@author			chorogyi@hallym.ac.kr
@warning		ȭ 
@return value	fdr => input file descriptor
***************************************************************/
void DBC_BS(int fdr)
{
	// local variable
	int temp = 0;
	int readLen = 0;
	int state = 0;
	
	int fdmw = open64("./READMETA", O_RDWR | O_TRUNC | O_CREAT);
	
	unsigned char sha1sum[20];
	unsigned char sha1sum_back[20];
	unsigned char buf[CHUNK_SIZE];
	unsigned char buf1[CHUNK_SIZE];
	unsigned char window1[CHUNK_SIZE+CHUNK_SIZE];
	
	u_int64_t filesize = lseek64(fdr, 0, SEEK_END);
	u_int64_t offset = 0;
	u_int64_t rabinf = 0;
	u_int64_t rabinp = 0;

	Chunk_Count = 0;
	
	lseek64(fdr, 0, SEEK_SET);
	
	myRabin.reset(); // RESET if you want to clear up the buffer in the RABINPOLY CLASS

	while(1)
	{
		temp = 0;
		memset(buf, '0', CHUNK_SIZE);

		state = VerifyFixedBlock(&offset, filesize, fdr, fdmw, buf, sha1sum, &rabinf, &Chunk_Count);

		if( state == -1 ) {
			break;
		}else if( state == 1 ) {
			continue;
		}

L1: //  ؽ
	
		temp = 0;
		memset(buf1, '0', CHUNK_SIZE); //  ʱȭ
		//   κ ˻
		if( offset + CHUNK_SIZE <= filesize )
		{
			read( fdr, buf1, CHUNK_SIZE);
			memcpy(window1, buf, CHUNK_SIZE);
			memcpy(window1+CHUNK_SIZE, buf1, CHUNK_SIZE);

			// window1  ޺κ ̵ ؽ ϰ  0  ߺ ߰
			if ( SlideVerify(window1, sha1sum_back, CHUNK_SIZE, &temp, &rabinp) < 0 )
			{
				if ( VerifyHash(window1, sha1sum, temp, &rabinf) > 0 ) // No duplication
				{
					InsertRabin(rabinf);
					unsigned char *sha1;
					sha1 = (unsigned char*)malloc(20);
					make_sha1_checksum(window1, temp, sha1);
					InsertSHA1(sha1);
					WriteLog(fdmw, rabinf, sha1, offset-CHUNK_SIZE, temp);
				}
				else // duplication
				{
					//WriteLog(fdmw, rabinf, sha1sum, offset-CHUNK_SIZE, temp);
					Chunk_Count += 1;
				}

				Chunk_Count += 1;
				offset = offset + temp + 1;
				//WriteLog(fdmw, rabinp, sha1sum_back, offset - CHUNK_SIZE, CHUNK_SIZE);
				lseek(fdr, offset, SEEK_SET);
				continue;
			}
			else
			{
				unsigned char *sha1;
				sha1 = (unsigned char*)malloc(20);
				
				if( rabinf == rabinp )
				{
					make_sha1_checksum(buf, CHUNK_SIZE, sha1sum);
					make_sha1_checksum(buf1, CHUNK_SIZE, sha1sum_back);

					if( HashComp((const void *)sha1sum, (const void *)sha1sum_back) == 0)
					{
						memcpy(buf, buf1, CHUNK_SIZE);
						Hashcpy((const void *)sha1, (const void *)sha1sum);
						InsertRabin(rabinf);
						InsertSHA1(sha1);
						rabinf = rabinp;
						WriteLog(fdmw, rabinf, sha1, offset-CHUNK_SIZE, CHUNK_SIZE);
						Chunk_Count += 1;
						lseek(fdr, offset, SEEK_SET);
						continue;
					}
				}

				make_sha1_checksum(buf, CHUNK_SIZE, sha1);
				InsertRabin(rabinf);
				InsertSHA1(sha1);
				WriteLog(fdmw, rabinf, sha1, offset-CHUNK_SIZE, CHUNK_SIZE);
				
				memcpy(buf, buf1, CHUNK_SIZE);
				rabinf = rabinp;
				offset += CHUNK_SIZE;
				goto L1;
			}
		}
		else
		{

			readLen = filesize - offset;
			read( fdr, buf1, readLen);
			memcpy(window1, buf, CHUNK_SIZE);
			memcpy(window1+CHUNK_SIZE, buf1, readLen);

			state = VerifyTailBlock(window1, buf, buf1, sha1sum_back, sha1sum, &rabinp, &rabinf, &offset, filesize, fdmw, fdr);

			if( state == -1 ) {
				break;
			}else if( state == 1 ) {
				continue;
			}
		}
	}
}

/***************************************************************
Main Function
***************************************************************/
/*
int main(int argc, char *argv[])
{
	int fdr, fdw;
	
	SHA1tree = RBTreeCreate(HashComp, HashDest, InfoDest, HashPrint, InfoPrint);
	Rabintree = RBTreeCreate(UInt64Comp, UInt64Dest, InfoDest, UInt64Print, InfoPrint);

	//initial time
	start.tv_usec = 0;
	start.tv_sec = 0;
	end.tv_usec = 0;
	end.tv_sec = 0;

	gettimeofday(&start, NULL);

	int c;
	int dedup_option = 0;
	int comp_option = 0;
	int stride_size = 0;
	
	extern char *optarg;	
	extern int optind, opterr, optopt;


	while ((c = getopt(argc, argv, "o:c:s:")) != -1) {		
		char *endp;		
		switch (c) {
		case 'o':
			if( strcmp(optarg, "fs") == 0)
			{
				dedup_option = 1;
			}
			else if( strcmp(optarg, "bs") == 0)
			{
				dedup_option = 2;
			}
			else if( strcmp(optarg, "ss") == 0)
			{
				dedup_option = 3;
			}
			else if( strcmp(optarg, "iss") == 0)
			{
				dedup_option = 4;
			}
			break;
		case 'c':
			if( strcmp(optarg, "zlib") == 0)
			{
				comp_option = 1;
			}
			else if( strcmp(optarg, "bz2") == 0)
			{
				comp_option = 2;
			}
			else if( strcmp(optarg, "lzo") == 0)
			{
				comp_option = 3;
			}
			break;
		case 's':
			stride_size = atoi(optarg);
			break;
		default:
			printf("dedup option => -o fs : Fixed Match, bss : byte shift!!, ss : stride scheme\n");
			printf("comp option => -c zlib bz2 lzo\n");
			printf("stride size => -s size\n");
			printf("./FD2E -o option -c comp -s size sourcefile destfile");
			exit(-1);
			break;				
		}
	}

	argc -= optind;	
	argv += optind;

	if( argc < 1 )
	{
		printf("dedup option => -o fs : Fixed Match, bss : byte shift!!, ss : stride scheme\n");
		printf("comp option => -c zlib bz2 lzo\n");
		printf("stride size => -s size\n");
		printf("./FD2E -o option -c comp -s size sourcefile destfile");
		exit(-1);
	}
	
	fdr = open64(*argv, O_RDONLY);
	fdw = open64("./RESULT", O_RDWR | O_TRUNC | O_CREAT);
	
	if( fdr < 0)
	{
		printf("File is not found\n");
		exit(-1);
	}

	printf("first pass....\n");
	// deduplication
	switch( dedup_option )
	{
		case 0 :
			switch( comp_option )
			{
				case 0 :
					;
					goto Lmain;
				case 1 :
					zlib_compress(fdr);
					goto Lmain;
				case 2 :
					bz2_compress(fdr);
					goto Lmain;
				case 3 :
					lzo_compress(fdr);
					goto Lmain;
			}
			break;
		case 1 :
			DBC_FS(fdr);
			break;
		case 2 :
			DBC_BS(fdr);
			break;
		case 3 :
			DBC_SS(fdr, stride_size);
			break;
		case 4 :
			DBC_ISS(fdr, stride_size);
			break;
	}


	printf("second pass....\n");
	// compress
	switch( comp_option )
	{
		case 0 :
			Plain(fdr, fdw);
			break;
		case 1 :
			Zlib_Blk_Comp(fdr, fdw);
			break;
		case 2 :
			Bz2_Blk_Comp(fdr, fdw);
			break;
		case 3 :
			Lzo_Blk_Comp(fdr, fdw);
			break;
	}

Lmain:
	
	close(fdr);
	close(fdw);
	printf("Duplication Chunk Count : %d, Block Count : %ld\n", Chunk_Count, SHA1tree->Count);	

	gettimeofday(&end, NULL);
	long checktime = (end.tv_sec*1000000 - start.tv_sec*1000000) + (end.tv_usec - start.tv_usec);
	printf("processtime: %ld milisec\n", checktime/1000);
	return 1;
	
}
*/