#include "Comp.h"

static HEAP_ALLOC(wrkmem,LZO1X_1_MEM_COMPRESS);

int lzo_compress(char *filename)
{
	void* addr_W;
	void* addr_R;
	int err;
	char addfilename[300];
	unsigned char out[CHUNK2];
	unsigned char in[CHUNK2];
	unsigned long offset_r, offset_w;
	u_int64_t filesize = 0LL;
	u_int64_t readsize = 0LL;
	u_int64_t writesize = 0LL;
	
	int r;
		
	sprintf(addfilename, "%s_lzofile", filename);
	
	int fdr = open64(filename, O_RDONLY);
	int fdw = open64("./RESULT", O_RDWR | O_TRUNC | O_CREAT);

	if( fdr < 0 || fdw < 0 ) {
		// error confirm
		printf("File open error\n");
		return(-200);
	}
	
	filesize = lseek64(fdr, 0, SEEK_END);
	lseek64(fdr, 0, SEEK_SET);
	
	// initialize the LZO library
	if( lzo_init() != LZO_E_OK)
	{
	    printf("internal error - lzo_init() failed !!!\n");
        printf("(this usually indicates a compiler bug - try recompiling\nwithout optimizations, and enable `-DLZO_DEBUG' for diagnostics)\n");
        return -3;
	}
	
	while( readsize < filesize )
	{
		memset(out, '\0', CHUNK2);
		if( readsize + CHUNK < filesize ) 
			offset_r = CHUNK;
		else
			offset_r = filesize - readsize;
		
		// zlib Compress
		addr_R = mmap64(0, offset_r, PROT_READ, MAP_SHARED, fdr, readsize);
		offset_w = CHUNK2;
		
		//lzo_memset(in, 0, offset_r);
		
		r = lzo1x_1_compress((const unsigned char*)addr_R, offset_r, out, &offset_w, wrkmem);
		
		if (r == LZO_E_OK)
		{
			; //printf("compressed %lu bytes into %lu bytes\n", (unsigned long) offset_r, (unsigned long) offset_w);
		}
		else
		{
			/* this should NEVER happen */
			printf("internal error - compression failed: %d\n", r);
			return -2;
		}
		
		/* check for an incompressible block */
		if (offset_w >= offset_r)
		{
			; //printf("This block contains incompressible data.\n");
			//return -1;
		}
				
		// writing file
		write(fdw, out, offset_w);
		writesize += offset_w;
		
		// memory de-llocate
		munmap(addr_R, offset_r);
		readsize += CHUNK;
	}
	
	return(1);
}

int zlib_compress(char *filename)
{
	void* addr_W;
	void* addr_R;
	int err;
	char addfilename[300];
	unsigned char out[CHUNK2];
	unsigned char in[CHUNK2];
	unsigned long offset_r, offset_w;
	u_int64_t filesize = 0LL;
	u_int64_t readsize = 0LL;
	u_int64_t writesize = 0LL;
		
	sprintf(addfilename, "%s_zlibfile", filename);
	
	int fdr = open64(filename, O_RDONLY);
	int fdw = open64("./RESULT", O_RDWR | O_TRUNC | O_CREAT);

	if( fdr < 0 || fdw < 0 ) {
		// error confirm
		printf("File open error\n");
		return(-200);
	}
	
	filesize = lseek64(fdr, 0, SEEK_END);
	lseek64(fdr, 0, SEEK_SET);
	
	while( readsize < filesize )
	{
		memset(out, '\0', CHUNK2);
		if( readsize + CHUNK < filesize ) 
			offset_r = CHUNK;
		else
			offset_r = filesize - readsize;
		
		// zlib Compress
		addr_R = mmap64(0, offset_r, PROT_READ, MAP_SHARED, fdr, readsize);
		offset_w = CHUNK2;
		err = compress(out, &offset_w, (const Bytef*)addr_R, offset_r);
		CHECK_ERR_RETURN(err, "compress");
				
		// writing file
		write(fdw, out, offset_w);
		writesize += offset_w;
		
		// memory de-llocate
		munmap(addr_R, offset_r);
		readsize += CHUNK;
	}
	
	return(1);
}

int bz2_compress(char *filename)
{
	void* addr_W;
	void* addr_R;
	int err;
	int BZ2;
	char addfilename[300];
	unsigned char out[CHUNK2];
	unsigned char in[CHUNK2];
	unsigned long offset_r, offset_w;
	u_int64_t filesize = 0LL;
	u_int64_t readsize = 0LL;
	u_int64_t writesize = 0LL;
		
	sprintf(addfilename, "%s_zlibfile", filename);
	
	int fdr = open64(filename, O_RDONLY);
	int fdw = open64("./RESULT", O_RDWR | O_TRUNC | O_CREAT);

	if( fdr < 0 || fdw < 0 ) {
		// error confirm
		printf("File open error\n");
		return(-200);
	}
	
	filesize = lseek64(fdr, 0, SEEK_END);
	lseek64(fdr, 0, SEEK_SET);

	while( readsize < filesize )
	{
		memset(out, '\0', CHUNK2);
		if( readsize + CHUNK < filesize ) 
			offset_r = CHUNK;
		else
			offset_r = filesize - readsize;
		
		// zlib Compress
		addr_R = mmap64(0, offset_r, PROT_READ, MAP_SHARED, fdr, readsize);
		offset_w = CHUNK2;

		BZ2 = BZ2_bzBuffToBuffCompress( out, &offset_w, (char *)addr_R, offset_r, 4, 0, 30);

		if( BZ2 == BZ_OUTBUFF_FULL)
		{
			memset(out, '\0', CHUNK2);
			memcpy(out, addr_R, offset_r);
		}
		
		// writing file
		write(fdw, out, offset_w);
		writesize += offset_w;
		
		// memory de-llocate
		munmap(addr_R, offset_r);
		readsize += CHUNK;
	}
	
	return(1);
}

